/*
* Licensed Materials - Property of IBM Corp.
* IBM UrbanCode Build
* IBM UrbanCode Deploy
* IBM UrbanCode Release
* IBM AnthillPro
* (c) Copyright IBM Corporation 2002, 2013. All Rights Reserved.
*
* U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by
* GSA ADP Schedule Contract with IBM Corp.
*/
package com.urbancode.anthill3

import groovy.util.slurpersupport.GPathResult

public class AHPTool {

    //**************************************************************************
    // CLASS
    //**************************************************************************

    static protected enum PropertyType {
        AGENT('Agent'),
        BUILD_LIFE('BuildLife'),
        JOB('Job'),
        REQUEST('Request'),
        STEP('Step'),
        SYSTEM('System'),
        ORIGINATING_REQUEST('OriginatingRequest'),
        ORIGINATING_WORKFLOW('OriginatingWorkflow')

        final def name
        PropertyType(def name) {
            this.name = name
        }
    }

    /**
     * Utility to manage child processes. Captures exitValue and allows for complete access to underlying ProcessBuilder
     * 1) will ensure child process is terminated if this process receives ctrl-C
     * 2) will ensure child process is terminated if any sort of exception is encountered handling of the child process
     * 3) will ensure child process IO streams are fully read+flushed
     * 4) defaults child out|err streams to this process out|err
     */
    static private class ProcUtil {
        ProcessBuilder pb = new ProcessBuilder([]);
        def inputText = null; // a Writable or String
        InputStream inputData = null;
        OutputStream out = System.out;
        OutputStream err = System.err;
        int exitValue;

        public ProcUtil(List<String> cmd) {
            // Use String[] to force any List GStrings elements to resolve
            this(cmd as String[])
        }

        public ProcUtil(String[] cmd) {
            pb.command(cmd);
        }

        public void execute() {
            Runtime runtime = Runtime.getRuntime();
            Process proc = null;
            Thread hook = null;
            try {
                // need to use shutdown hooks in order to cascade sig-int (ctrl-C) to child proc
                hook = { proc?.destroy();} as Thread;
                runtime.addShutdownHook(hook);

                proc = pb.start();

                Thread errThread = proc.consumeProcessErrorStream(err);
                Thread outThread = proc.consumeProcessOutputStream(out);
                if (inputText != null) {
                    // copy textual data to stdin
                    proc.withOutputStream{ procStdIn ->
                        new OutputStreamWriter(procStdIn, 'utf-8').withWriter{ it << inputText }
                    }
                }
                else if (inputData != null) {
                    // copy binary data to stdin
                    proc.withOutputStream{ procStdIn ->
                        procStdIn << inputData
                    }
                }
                else {
                    // no data passed to stdin
                    proc.out.close();
                }
                exitValue = proc.waitFor()
                errThread.join(5000)
                outThread.join(1000)
            }
            finally {
                // best effort to kill the child process if this process is aborted
                if (proc != null) {
                    proc.destroy();
                }
                if (hook != null) {
                    // cleanup non-shutdown scenarios
                    runtime.removeShutdownHook(hook);
                }
            }
        }
    }

    //**************************************************************************
    // INSTANCE
    //**************************************************************************

    final def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
    final def ahptool = isWindows ? 'ahptool.cmd' : 'ahptool'
    def out = System.out;
    def err = System.err;

    public AHPTool(){
    }

    public void addChangeSets(changeSetXML) {
        ProcUtil proc = new ProcUtil([ahptool, 'addSourceChanges', '-'])
        proc.inputText = changeSetXML;
        proc.execute();
        if (proc.exitValue) {
            throw new Exception("Failed to add Change Sets using ahptool (exit=" + proc.exitValue + "): " + changeSetXML)
        }
    }

    /**
     * Get a map of bug id to changeset comment.
     *
     * @param regex the regex to match comments against
     * @return a map of bugid to changeset comment
     */
    public Map<String,String> getChangeSets(String regex) {
        ProcUtil proc = new ProcUtil([ahptool, "getSourceChanges"])
        proc.out = new ByteArrayOutputStream();
        proc.execute();
        def changeSetXML = proc.out.toString('utf-8')
        if (proc.exitValue) {
            throw new Exception("Failed to get source changes from ahptool (exit=" + proc.exitValue + "): $changeSetXML")
        }
        def pattern = java.util.regex.Pattern.compile(regex)
        def Map<String, String> id2changecomment = new HashMap<String,String>()

        // will throw a parse exception at random times - any ideas?
        new XmlSlurper().parseText(changeSetXML)."change-set".each { element ->
            def matcher = pattern.matcher(element.comment.text())
            while (matcher.find()) {
                def patternid
                def comment
                if (matcher.groupCount() > 0) {
                    // they specified a '(...)' group within the pattern, use that as the bug-id
                    patternid = matcher.group(1)
                }
                else {
                    // use the whole matching substring as the bug-id
                    patternid = matcher.group()
                }
                comment = (element."comment".text())
                if (patternid != null && comment != null) {
                    id2changecomment.put(patternid,comment)
                }
            } // end finding matches
        } // end for each change-set

        return id2changecomment
    }

    /**
     * Get the raw changeset xml data
     * @return
     */
    public String getChangeSetsXml() {
        ProcUtil proc = new ProcUtil([ahptool, "getSourceChanges"])
        proc.out = new ByteArrayOutputStream();
        proc.execute();
        def changeSetXML = proc.out.toString('utf-8');
        if (proc.exitValue) {
            throw new Exception("Failed to get source changes from ahptool (exit=" + proc.exitValue + "): $changeSetXML")
        }
        return changeSetXML
    }

    /**
     * Get the dependency plan xml for a given build life
     * @return
     */
    public String getDependencyPlanXml(String buildLifeId) {
        ProcUtil proc = new ProcUtil([ahptool, "getDependencyPlan", "-buildlife", buildLifeId]);
        proc.out = new ByteArrayOutputStream();
        proc.execute();
        def depXml = proc.out.toString('utf-8');
        if (proc.exitValue) {
            throw new Exception("Failed to get dependency plan from ahptool (exit=" + proc.exitValue + "): " + depXml)
        }
        return depXml;
    }

    /**
     * Get a map of bugid to changeset
     * @param regex the regex to match comments against
     * @return a map of bugid to changeset element
     */
    public Map<String,GPathResult> getMatchingChangeSets(String regex) {
        ProcUtil proc = new ProcUtil([ahptool, "getSourceChanges"])
        proc.out = new ByteArrayOutputStream();
        proc.execute();
        def changeSetXML = proc.out.toString('utf-8');
        if (proc.exitValue) {
            throw new Exception("Failed to get source changes from ahptool (exit=" + proc.exitValue + "): $changeSetXML")
        }
        def pattern = java.util.regex.Pattern.compile(regex)
        def Map<String, GPathResult> id2changeSet = new HashMap<String,GPathResult>()

        // will throw a parse exception at random times - any ideas?
        new XmlSlurper().parseText(changeSetXML)."change-set".each { element ->
            def matcher = pattern.matcher(element.comment.text())
            while (matcher.find()) {
                def patternid
                def comment
                if (matcher.groupCount() > 0) {
                    // they specified a '(...)' group within the pattern, use that as the bug-id
                    patternid = matcher.group(1)
                }
                else {
                    // use the whole matching substring as the bug-id
                    patternid = matcher.group()
                }
                if (patternid != null) {
                    id2changeSet.put(patternid,element)
                }
            } // end finding matches
        } // end for each change-set

        return id2changeSet
    }

    public void addReports(reportName, data) {
        ProcUtil proc = new ProcUtil([ahptool, '-report', reportName, 'addReports', '-'])
        proc.inputData = data
        proc.execute();
        if (proc.exitValue) {
            throw new Exception("Failed to add report using ahptool (exit=" + proc.exitValue + "): " + reportName)
        }
    }

    public void addSourceAnalytics(reportName, sourceAnalyticsXml) {
        ProcUtil proc = new ProcUtil([ahptool, '-report', reportName, 'addSourceAnalytics', '-'])
        proc.inputText = sourceAnalyticsXml;
        proc.execute();
        if (proc.exitValue) {
            throw new Exception("Failed to add Source Analytics using ahptool (exit=" + proc.exitValue + "): " + sourceAnalyticsXml)
        }
    }

    public void addTests(reportName, testsXml) {
        ProcUtil proc = new ProcUtil([ahptool, '-report', reportName, 'addTests', '-'])
        proc.inputText = testsXml;
        proc.execute();
        if (proc.exitValue) {
            throw new Exception("Failed to add Test Results using ahptool (exit=" + proc.exitValue + "): " + testsXml)
        }
    }

    /**
     *
     * @param issuesXml a Writable UTF-8 matching the issue schema
     */
    public void addIssues(issuesXml) {
        ProcUtil proc = new ProcUtil([ahptool, "addIssues", '-'])
        proc.inputText = issuesXml;
        proc.execute();
        if (proc.exitValue) {
            throw new Exception("Failed to add Issues using ahptool (exit=" + proc.exitValue + "): " + issuesXml)
        }
    }

    public String getIssuesXml() {
        // get the changes of the build life from ahptool
        ProcUtil proc = new ProcUtil([ahptool, "getIssues"])
        proc.out = new ByteArrayOutputStream();
        proc.execute();
        def issuesXml = proc.out.toString('utf-8');
        if (proc.exitValue) {
            throw new Exception("Failed to get build life issues from ahptool (exit=" + proc.exitValue + "): " + issuesXml)
        }
        return issuesXml;
    }

    public Map<String,String> getAgentProperties() {
        return getProperties(AHPTool.PropertyType.AGENT)
    }


    /**
     * @since 3.8 (2011-04-11)
     * @param linksXml a Writable or String UTF-8 matching the buildlife links schema
     */
    public void addBuildLifeLinks(def linksXml) {
        executeWithXml('addBuildLifeLinks', linksXml)
    }

    /**
     * @since 3.8 (2011-04-11)
     * @return xml document of the build life links
     */
    public String getBuildLifeLinks() {
        return execute("getBuildLifeLinks");
    }

    public Map<String,String> getBuildLifeProperties() {
        return getProperties(AHPTool.PropertyType.BUILD_LIFE)
    }

    public Map<String,String> getBuildLifeProperties(def id) {
        ProcUtil proc = new ProcUtil([ahptool, "getBuildLifeProperties", '-buildlife', id])
        proc.out = new ByteArrayOutputStream();
        proc.execute();
        def propertiesXml = proc.out.toString('utf-8');
        if (proc.exitValue) {
            throw new Exception("Failed to get BuildLife ${id} properties from ahptool (exit=" + proc.exitValue + "): " + propertiesXml)
        }
        def props = [:]
        new XmlSlurper().parseText(propertiesXml).property.each{ propElem ->
            props[propElem.@name.text()] = propElem.text()
        }
        return props
    }

    public Map<String,String> getOriginatingRequestProperties() {
        return getProperties(AHPTool.PropertyType.ORIGINATING_REQUEST)
    }

    public Map<String,String> getOriginatingWorkflowProperties() {
        return getProperties(AHPTool.PropertyType.ORIGINATING_WORKFLOW)
    }

    public Map<String, String> getJobProperties() {
        return getProperties(AHPTool.PropertyType.JOB)
    }

    public Map<String, String> getRequestProperties() {
        return getProperties(AHPTool.PropertyType.REQUEST)
    }

    public Map<String,String> getStepProperties() {
        return getProperties(AHPTool.PropertyType.STEP)
    }

    public Map<String,String> getSystemProperties() {
        return getProperties(AHPTool.PropertyType.SYSTEM)
    }

    public Map<String, String> getProperties(AHPTool.PropertyType propType) {
        ProcUtil proc = new ProcUtil([ahptool, "get${propType.name}Properties"])
        proc.out = new ByteArrayOutputStream();
        proc.execute();
        def propertiesXml = proc.out.toString('utf-8');
        if (proc.exitValue) {
            throw new Exception("Failed to get ${propType.name} properties from ahptool (exit=" + proc.exitValue + "): " + propertiesXml)
        }
        def props = [:]
        new XmlSlurper().parseText(propertiesXml).property.each{ propElem ->
            props[propElem.@name.text()] = propElem.text()
        }
        return props
    }

    public StepInfo getStepInfo() {
        ProcUtil proc = new ProcUtil([ahptool, "getStepInfo"]);
        proc.out = new ByteArrayOutputStream();
        proc.execute();
        if (proc.exitValue) {
            throw new Exception("Failed to get step info from ahptool (exit=" + proc.exitValue + "): " + proc.out.toString('utf-8'));
        }
        String infoXml = proc.out.toString('utf-8');
        StepInfo result = null;
        try {
            result = new StepInfo(new XmlSlurper().parseText(infoXml));
        }
        catch (Exception e) {
            throw new Exception("Failed to parse step info from ahptool: " + infoXml);
        }
        return result;
    }

    public String callGetCommand(def commandName, def args) {
        def command = [ahptool, commandName]
        command.addAll(args)
        ProcUtil proc = new ProcUtil(command);
        proc.out = new ByteArrayOutputStream();
        proc.execute();
        String output = proc.out.toString('utf-8');
        if (proc.exitValue) {
            throw new Exception("Failed executing $commandName with ahptool [$command] (exit=" + proc.exitValue + "): " + output)
        }
        return output;
    }

    public void callSetCommand(def commandName, def args, def input) {
        def command = [ahptool, commandName]
        command.addAll(args)
        command.add("-")
        ProcUtil proc = new ProcUtil(command);
        proc.inputText = input;
        proc.out = new ByteArrayOutputStream();
        proc.execute();
        if (proc.exitValue) {
            throw new Exception("Failed executing $commandName with ahptool [$command] (exit=" + proc.exitValue + "): " + proc.out.toString('utf-8'))
        }
    }

    public void setAgentProperties(def props) {
        setProperties(props, AHPTool.PropertyType.AGENT)
    }

    public void setBuildLifeProperties(def props) {
        setProperties(props, AHPTool.PropertyType.BUILD_LIFE)
    }

    public void setJobProperties(def props) {
        setProperties(props, AHPTool.PropertyType.JOB)
    }

    public void setRequestProperties(def props) {
        setProperties(props, AHPTool.PropertyType.REQUEST)
    }

    public void setStepProperties(def props) {
        setProperties(props, AHPTool.PropertyType.STEP)
    }

    public void setSystemProperties(def props) {
        setProperties(props, AHPTool.PropertyType.SYSTEM)
    }

    protected void setProperties(def props, def propType) {
        def propXml = groovy.xml.DOMBuilder.newInstance().'properties'{
            props.each{key, value ->
                'property'('name':key, value)
            }
        }

        def uploadChangeDateCmd = [ahptool, "set${propType.name}Properties", '-']
        ProcUtil proc = new ProcUtil(uploadChangeDateCmd);
        proc.inputText = propXml;
        proc.execute();
        if (proc.exitValue) {
            throw new Exception("Failed to set $propType.name properties using ahptool (exit=" + proc.exitValue + "): " + propXml)
        }
        println "";
    }

    public void setQuietPeriodLatestTime(long unixTimeMillis) {
        setStepProperties(['latest-change':unixTimeMillis]);
    }

    public void setQuietPeriodLatestTime(Date date) {
        setQuietPeriodLatestTime(date.time);
    }

    public void addTestCoverage(reportName, testCoverageXml) {
        ProcUtil proc = new ProcUtil([ahptool, '-report', reportName, 'addTestCoverage', '-']);
        proc.inputText = testCoverageXml;
        proc.execute();
        if (proc.exitValue) {
            throw new Exception("Failed to add Code Coverage using ahptool (exit=" + proc.exitValue + "): " + testCoverageXml)
        }
    }

    protected String execute(String commandName) {
        ProcUtil proc = new ProcUtil([ahptool, commandName]);
        proc.out = new ByteArrayOutputStream();
        proc.execute();
        if (proc.exitValue) {
            throw new Exception("Failed to $commandName using ahptool (exit=" + proc.exitValue + "): ")
        }
        return proc.out.toString('utf-8');
    }

    protected executeWithXml(String commandName, def xml) {
        ProcUtil proc = new ProcUtil([ahptool, commandName, '-']);
        proc.inputText = xml;
        proc.execute();
        if (proc.exitValue) {
            throw new Exception("Failed to $commandName using ahptool (exit=" + proc.exitValue + "): " + xml)
        }
    }

    public String getWorkDir() {
        return execute('getWorkDir')
    }

    public setWorkDir(String workDir, boolean lockForWorkflow, boolean releasePrior) {
        def lockScope = lockForWorkflow ? 'worklow' : 'job'
        def workDirXml = groovy.xml.DOMBuilder.newInstance().'work-dir'('scope':lockScope, 'releasePrior':releasePrior, targetWorkDir)
        executeWithXml('setWorkDir', workDirXml)
    }
}
