/*
 * Inputs:
 *     Set<PluginSourceConfig> availableSourceConfigSet : the source configurations which should be scanned
 *     Map<String, String> triggerPropertyMap : parameters passed to the trigger invocation
 *     Map<String, String> requestPropertyMap: parameters to be placed upon the build request for matching source-configs
 *     Logger log : a logger instance
 *
 * Output:
 *     Set<PluginSourceConfig> matchingSourceConfigSet : an empty set used to return source configurations which should trigger
 *
 * Return:
 *     nothing
 */
import com.urbancode.anthill3.domain.property.PropertyValue;
import com.urbancode.anthill3.domain.property.PropertyValueGroupFactory;
import groovy.json.JsonSlurper;

String.metaClass.endsWithIgnoreCase = { String suffix ->
    int dl = delegate.length();
    int sl = suffix.length();
    return dl >= sl && delegate.substring(dl - sl, dl).equalsIgnoreCase(suffix);
};

// normalize URI-like strings
String.metaClass.normalize = {
    return delegate.replaceAll('(?<!:)//+', '/');
};

public class GitRepoSource {

    // the hostname of the git repo server
    String host = '';
    // the path portion of the git repo, does not include "git" extension
    String path = '';
    String branch = '';

    public GitRepoSource() {
    }

    public GitRepoSource(String url) {
        url = url.normalize();

        // scp style URL
        def scpParser = ~'^(?:[a-zA-Z0-9]+@)?([^:]+):(.*)'

        // ssh/git/http protocol style URL
        def protocolParser = ~'^[a-zA-Z]+://(?:[^/@]+@)?([^/:]+)(?::\\d*)?(?:/(.*)|$)'

        def mProto = (url =~ protocolParser)
        def mScp = (url =~ scpParser)

        if (mProto.matches()) {
            this.setHost(mProto.group(1));
            this.setPath(mProto.group(2));
        }
        else if (mScp.matches()) {
            this.setHost(mScp.group(1));
            this.setPath(mScp.group(2));
        }
        else {
            this.setPath(url);
        }
    }


    public void setPath(String path) {
        // normalize path and remove any trailing .git and/or /
        this.path = path?.normalize()?.replaceAll('(\\.git)?/?$', '');
    }

    public void setBranch(String branch) {
        // remove any refs/heads prefix on ref
        this.branch = branch.replaceAll('refs/heads/', '');
    }

    /**
     * Append a page-relative repo-source to this repo-source
     *
     * @param child a page-relative repo-source
     * @return
     */
    public GitRepoSource plus(GitRepoSource child) {
        if (child.host) {
            return child;
        }

        GitRepoSource result = new GitRepoSource();
        result.host = this.host;
        if (this.path) {
            result.path = this.path + '/';
        }
        else {
            result.path = '';
        }
        result.path = result.path + child.path;

        result.branch = child.branch;

        return result;
    }

    public boolean matches(GitRepoSource repoSource) {
        // blank host is treated as a wildcard
        return (this.host == null || repoSource.host == null|| this.host == repoSource.host) &&
            repoSource.path == path &&
            repoSource.branch == branch;

    }

    public String toString() {
        return "host:${this.host},path:${this.path},branch:${this.branch}"
    }
}

def parseLong = { value ->
   return value ? value.toLong() : null
}

def resolveValue = {def module, def value ->
    return ParameterResolver.resolveForSourceConfig(value, module.getContainer());
}

def resolveModuleProperty = {def module, def propertyName ->
    return resolveValue(module, module.getPropertyValue(propertyName)?.value);
}

/**
 * if remoteUrl is relative, concatinate it with the base, otherwise use that
 * also normalize the url and strip trailing .git and/or slash
 */
def calculateRepositorySource = { String remoteBaseUrl, String remoteUrl ->
    GitRepoSource result = new GitRepoSource(remoteUrl);
    if (!result.host) {
        result = new GitRepoSource(remoteBaseUrl) + result;
    }
    return result;
}

def buildSourcePropertyMap = { module ->
    def result = [:]
    module.getPropertyMap().each{ key, value ->
        result[key] = value?.value
    }

    Long scRepoId = parseLong(result['repo']);
    if (scRepoId) {
        def pvgFactory = PropertyValueGroupFactory.getInstance();
        def repoPvg = pvgFactory.restore(scRepoId);
        result['repo'] = repoPvg.name;
        repoPvg.getPropertyMap().each{ key, value ->
            result['repo/'+key] = value?.value
        }
    }
    return result;
}

//
//
//

String revision = triggerPropertyMap['revision'];
String fetchRef = triggerPropertyMap['fetchRef'];
String postData = triggerPropertyMap['postData'];

GitRepoSource triggeredRepoSource = null;



// our "normal" git hook post data
String triggerRepoPath = triggerPropertyMap['repo'];
if (triggerRepoPath) {
    revision = triggerPropertyMap['revision']
    String triggerRef   = triggerPropertyMap['branch'];

    triggeredRepoSource = new GitRepoSource(triggerRepoPath);
    if (!triggeredRepoSource.host && triggeredRepoSource.path.contains("/")) {
        // handle legacy cases where a protocol-less url which includes hostname was passed to trigger
        def split = triggeredRepoSource.path.split("/", 2);
        triggeredRepoSource.host = split[0];
        triggeredRepoSource.path = split[1];
    }
    triggeredRepoSource.branch = triggerRef;
}



// github + gitlab  webhook compatibility
String gitHubPayload = triggerPropertyMap['payload'];
if (gitHubPayload) {
    def gson = new com.google.gson.GsonBuilder().create();
    def data = gson.fromJson(gitHubPayload, Map.class);

    String triggerRef = data.ref
    triggeredRepoSource = new GitRepoSource(data.repository.url);
    triggeredRepoSource.branch = triggerRef?.split('[/\\\\]')?.last();
    revision = data.after
}
else if (postData) {
    // gitlab compatibility
    def gitlabPayload = new JsonSlurper().parseText(postData);
    if (gitlabPayload.ref && gitlabPayload.repository.url) {
        String triggerRef = gitlabPayload.ref;
        triggeredRepoSource = new GitRepoSource(gitlabPayload.repository.url);
        triggeredRepoSource.branch = triggerRef?.split('[/\\\\]')?.last();
        revision = gitlabPayload.after
    }
}


log.debug("Trigger notified of changes in repo: $triggeredRepoSource")

if (triggeredRepoSource && triggeredRepoSource.branch) {

    for (sc in availableSourceConfigSet) {
        for (module in sc.getSourcePropertyValueGroups()) {
            def plugin = module.plugin;
            if (plugin.pluginId.endsWith('.plugin.Git')) {
                Map<String,String> scProps = buildSourcePropertyMap(module);

                String scRepoBasePath = scProps['repo/repoBaseUrl'];
                String scRepoPath = resolveModuleProperty(module, "remoteUrl")?.replaceAll(~'[/\\\\]+$', '');
                String scBranch = resolveModuleProperty(module, "branch") ?: 'master';

                GitRepoSource source = calculateRepositorySource(scRepoBasePath, scRepoPath);
                source.branch = scBranch;
                log.debug("parsed $scRepoBasePath + $scRepoPath of into $source")
                if (source) {
                    if (triggeredRepoSource.matches(source)) {
                        if (revision) {
                            requestPropertyMap["workspace.revision.${module.name}"] = new PropertyValue(revision, false);
                        }
                        if (fetchRef) {
                            requestPropertyMap["workspace.ref.${module.name}"] = new PropertyValue(fetchRef, false);
                        }
                        matchingSourceConfigSet.add(sc)
                        log.debug("Found matching source config from project $sc.project.name with values $source")
                    }
                    else {
                        log.debug("Trigger did not match source config from project $sc.project.name with values $source")
                    }
                }
            }
        }
    }
}
else {
    log.warn("Could not process Git repo trigger data: $triggerPropertyMap - invalid repo and/or branch parameters")
}
